from __future__ import annotations

import datetime
import uuid
from typing import Literal

import pytest

import typedmongo as mongo
from typedmongo.expressions import Expression


class MongoDocument(mongo.Document):
    __abstract__ = True

    _id: mongo.StringField = mongo.StringField(default=lambda: uuid.uuid4().hex)


class Wallet(mongo.Document):
    balance: mongo.DecimalField


class Social(mongo.Document):
    site: mongo.StringField
    user: mongo.StringField


class User(MongoDocument):
    name: mongo.StringField
    gender: mongo.LiteralField[Literal["m", "f"]]
    age: mongo.IntegerField
    tags: mongo.ListField[str]
    wallet: mongo.EmbeddedField[Wallet]
    created_at: mongo.DateTimeField = mongo.DateTimeField(
        default=lambda: datetime.datetime.now(datetime.timezone.utc)
    )
    children: mongo.ListField[User]
    socials: mongo.ListField[Social] = mongo.ListField(
        mongo.EmbeddedField(Social), default=list
    )
    extra: mongo.DictField = mongo.DictField(default=dict)


User.__lazy_init_fields__()


@pytest.mark.parametrize(
    "expression, repr_str",
    [
        (
            User.name == "Aber",
            "CompareExpression(field=name, operator='==', arg='Aber')",
        ),
        (
            User.age >= 18,
            "CompareExpression(field=age, operator='>=', arg=18)",
        ),
        (
            User.tags == "a",
            "CompareExpression(field=tags, operator='==', arg='a')",
        ),
        (
            User.tags[0] == "0",
            "CompareExpression(field=tags.0, operator='==', arg='0')",
        ),
        (
            User.wallet._.balance > 1,
            "CompareExpression(field=wallet.balance, operator='>', arg=1)",
        ),
        (
            User.children[0].name == "Yue",
            "CompareExpression(field=children.0.name, operator='==', arg='Yue')",
        ),
        (
            User.children._.age >= 18,
            "CompareExpression(field=children.age, operator='>=', arg=18)",
        ),
        (
            User.extra == {"a": "b"},
            "CompareExpression(field=extra, operator='==', arg={'a': 'b'})",
        ),
    ],
)
def test_expression(expression, repr_str):
    assert isinstance(expression, Expression)
    assert repr(expression) == repr_str


def test_list_field():
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
            "socials": [{"site": "github.com", "user": "abersheeran"}],
        }
    )
    assert user.socials == [Social(site="github.com", user="abersheeran")]


def test_field_default():
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
        }
    )
    assert isinstance(user._id, str)
    assert isinstance(user.created_at, datetime.datetime)

    user = User.load(
        {
            "name": "Aber",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
        },
        partial=True,
    )
    assert isinstance(user._id, str)
    assert isinstance(user.created_at, datetime.datetime)

    user = User(
        name="Aber",
        gender="m",
        age=18,
        tags=["a", "b"],
        wallet=Wallet(balance=100),
        children=[],
    )
    assert hasattr(user, "_id")
    assert isinstance(User.dump(user)["_id"], str)
    assert isinstance(user.created_at, datetime.datetime)


def test_recursion_field():
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [
                {
                    "name": "Yue",
                    "gender": "f",
                    "age": 18,
                    "tags": ["a", "b"],
                    "wallet": {"balance": 100},
                    "children": [],
                }
            ],
        }
    )
    assert isinstance(user.wallet, Wallet)
    assert user.wallet.balance == 100
    assert isinstance(user.children[0], User)
    assert user.children[0].name == "Yue"
    assert isinstance(user.children[0]._id, str)


def test_empty_field():
    user = User.load({}, partial=True)
    assert not hasattr(user, "name")
    assert hasattr(user, "_id")


def test_dict_field():
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
            "extra": {"a": "b"},
        }
    )
    assert user.extra == {"a": "b"}
    assert user.dump()["extra"] == {"a": "b"}


def test_datetime_field():
    user = User.load(dict(created_at=datetime.datetime.now()), partial=True)
    assert isinstance(user.created_at, datetime.datetime)


def test_literal_field():
    user = User.load(dict(gender="m"), partial=True)
    assert user.gender == "m"


class UserWithRole(User):
    role: mongo.LiteralField[Literal["admin", "user"]]


UserWithRole.__lazy_init_fields__()


def test_three_level_inheritance():
    user = UserWithRole.load(dict(role="admin"), partial=True)
    assert isinstance(user._id, str)



import pytest
from bson import ObjectId
from marshmallow import ValidationError

import typedmongo.asyncio as mongo


class MongoDocument(mongo.Document):
    _id: mongo.ObjectIdField


MongoDocument.__lazy_init_fields__()


def test_objectid():
    assert MongoDocument.load({"_id": "0123456789ab0123456789ab"})._id == ObjectId(
        "0123456789ab0123456789ab"
    )
    assert MongoDocument.load({"_id": b"foo-bar-quux"})._id == ObjectId(
        "666f6f2d6261722d71757578"
    )


def test_invalid_objectid():
    with pytest.raises(ValidationError):
        MongoDocument.load({"_id": "123"})



from __future__ import annotations

import dataclasses
import decimal
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Optional,
    TypeVar,
    get_args,
    get_origin,
    overload,
)

from bson import ObjectId
from marshmallow import fields
from typing_extensions import Self

from typedmongo.expressions import CompareMixin, HasFieldName, OrderByMixin
from typedmongo.marshamallow import (
    MarshamallowDateTime,
    MarshamallowLiteral,
    MarshamallowObjectId,
)

if TYPE_CHECKING:
    from .table import Document

TypeDocument = TypeVar("TypeDocument", bound=type["Document"])
T = TypeVar("T", bound="Document")
TypeDocumentOrAny = TypeVar("TypeDocumentOrAny", bound=type["Document"] | Any)
FieldType = TypeVar("FieldType")


@dataclasses.dataclass
class FieldParamters(Generic[FieldType]):
    default: Optional[FieldType | Callable[[], FieldType]] = dataclasses.field(
        default=None, kw_only=True
    )


@dataclasses.dataclass(eq=False, order=False, unsafe_hash=True)
class Field(Generic[FieldType], OrderByMixin, CompareMixin):
    """
    Field
    """

    default: Optional[FieldType | Callable[[], FieldType]] = dataclasses.field(
        default=None, kw_only=True
    )
    field_name: str = dataclasses.field(init=False)
    marshamallow: fields.Field = dataclasses.field(init=False)

    def __set_name__(self, owner: type[Document], name: str) -> None:
        self._table = owner
        self._name = name

        self.field_name = name

        if not hasattr(self, "marshamallow"):
            self.marshamallow = fields.Field(required=True, allow_none=False)

        if self.default is not None:
            # https://github.com/marshmallow-code/marshmallow/issues/2151
            self.marshamallow.required = False
            self.marshamallow.load_default = self.default
            self.marshamallow.dump_default = self.default

    @overload
    def __get__(self: Self, instance: None, cls: type) -> Self: ...

    @overload
    def __get__(self: Self, instance: object, cls: type) -> FieldType: ...

    def __get__(self, instance, cls):
        if instance is None:  # Call from class
            return self

        try:
            return instance.__dict__[self._name]
        except KeyError:
            message = "{0} has no attribute '{1}'".format(instance, self._name)
            raise AttributeError(message) from None

    def __set__(self, instance: Document, value: Any) -> None:
        instance.__dict__[self._name] = value

    def __delete__(self, instance: Document) -> None:
        try:
            del instance.__dict__[self._name]
        except KeyError:
            message = "{0} has no attribute '{1}'".format(instance, self._name)
            raise AttributeError(message)

    def get_field_type(self) -> type[FieldType]:
        if hasattr(self, "__field_type__"):
            return self.__field_type__  # type: ignore
        for origin_base in self.__orig_bases__:  # type: ignore
            origin_class = get_origin(origin_base)
            if isinstance(origin_class, type) and issubclass(origin_class, Field):
                self.__field_type__ = generic_type = get_args(origin_base)[0]
                return generic_type
        raise RuntimeError(f"Cannot get field type for {self}")

    def load(self, value: Any, *, partial: bool = False) -> FieldType:
        return value

    def dump(self, value: FieldType) -> Any:
        return value

    def to_mongo(self, value: FieldType) -> Any:
        return value


@dataclasses.dataclass(eq=False)
class ObjectIdField(Field[ObjectId]):
    """
    ObjectId field
    """

    marshamallow: MarshamallowObjectId = dataclasses.field(
        default_factory=lambda: MarshamallowObjectId(required=True, allow_none=False)
    )


@dataclasses.dataclass(eq=False)
class LiteralField(Field[FieldType]):
    """
    Literal field
    """

    literal: type[FieldType]

    def __post_init__(self):
        self.marshamallow = MarshamallowLiteral(
            self.literal, required=True, allow_none=False
        )


@dataclasses.dataclass(eq=False)
class StringField(Field[str]):
    """
    String field
    """

    marshamallow: fields.String = dataclasses.field(
        default_factory=lambda: fields.String(required=True, allow_none=False)
    )


@dataclasses.dataclass(eq=False)
class IntegerField(Field[int]):
    """
    Integer field
    """

    marshamallow: fields.Integer = dataclasses.field(
        default_factory=lambda: fields.Integer(required=True, allow_none=False)
    )


@dataclasses.dataclass(eq=False)
class FloatField(Field[float]):
    """
    Float field
    """

    marshamallow: fields.Float = dataclasses.field(
        default_factory=lambda: fields.Float(required=True, allow_none=False)
    )


@dataclasses.dataclass(eq=False)
class BooleanField(Field[bool]):
    """
    Boolean field
    """

    marshamallow: fields.Boolean = dataclasses.field(
        default_factory=lambda: fields.Boolean(required=True, allow_none=False)
    )


@dataclasses.dataclass(eq=False)
class DateTimeField(Field[datetime]):
    """
    DateTime field
    """

    marshamallow: MarshamallowDateTime = dataclasses.field(
        default_factory=lambda: MarshamallowDateTime(required=True, allow_none=False)
    )


@dataclasses.dataclass(eq=False)
class DecimalField(Field[decimal.Decimal]):
    """
    Decimal field
    """

    marshamallow: fields.Decimal = dataclasses.field(
        default_factory=lambda: fields.Decimal(required=True, allow_none=False)
    )


@dataclasses.dataclass(eq=False)
class DictField(Field[dict]):
    """
    Dict field
    """

    marshamallow: fields.Dict = dataclasses.field(
        default_factory=lambda: fields.Dict(required=True, allow_none=False)
    )


@dataclasses.dataclass
class FieldNameProxy(Generic[TypeDocument]):
    prefix: HasFieldName
    t: TypeDocument

    def __get__(self, instance, owner) -> TypeDocument: ...

    def __getattr__(self, name: str) -> FieldNameProxyString:
        try:
            return FieldNameProxyString(
                f"{self.prefix.field_name}.{self.t.__fields__[name].field_name}"
            )
        except KeyError:
            message = "{0} has no attribute '{1}'".format(self.t, name)
            raise AttributeError(message) from None


@dataclasses.dataclass(eq=False)
class FieldNameProxyString(OrderByMixin, CompareMixin):
    field_name: str


@dataclasses.dataclass(eq=False)
class EmbeddedField(Generic[T], Field[T]):
    """
    Embedded field
    """

    _: FieldNameProxy[type[T]] = dataclasses.field(init=False)

    schema: type[T]

    def __post_init__(self):
        self._ = FieldNameProxy(self, self.schema)
        self.marshamallow = fields.Nested(lambda: self.schema.__schema__)

        def load(value: Any, *, partial: bool = False) -> T:
            return self.schema.load(value, partial=partial)

        def dump(value: T) -> dict[str, Any]:
            return self.schema.dump(value)

        def to_mongo(value: T) -> dict[str, Any]:
            return self.schema.to_mongo(value)

        self.load = load
        self.dump = dump
        self.to_mongo = to_mongo

    def __set_name__(self, owner: type[Document], name: str) -> None:
        if not issubclass(self.schema, owner):
            self.schema.__lazy_init_fields__()
        return super().__set_name__(owner, name)

    def get_field_type(self) -> type[T]:
        return self.schema


@dataclasses.dataclass(eq=False)
class ListFieldNameProxy(Generic[TypeDocumentOrAny], OrderByMixin, CompareMixin):
    number: int | None
    prefix: HasFieldName
    t: TypeDocumentOrAny

    @property
    def field_name(self) -> str:
        if self.number is None:
            return self.prefix.field_name
        return f"{self.prefix.field_name}.{self.number}"

    def __get__(self, instance, owner) -> TypeDocumentOrAny: ...

    def __getattr__(self, name: str) -> FieldNameProxyString:
        try:
            return FieldNameProxyString(
                f"{self.field_name}.{self.t.__fields__[name].field_name}"
            )
        except KeyError:
            message = "{0} has no attribute '{1}'".format(self.t, name)
            raise AttributeError(message) from None


@dataclasses.dataclass(eq=False)
class ListField(Generic[FieldType], Field[list[FieldType]]):
    """
    List field
    """

    _: ListFieldNameProxy[type[FieldType]] = dataclasses.field(init=False)

    field: Field[FieldType]

    def __getitem__(self, index: int) -> type[FieldType]:
        return ListFieldNameProxy(index, self, self.field.get_field_type())  # type: ignore

    def __post_init__(self):
        self._ = ListFieldNameProxy(None, self, self.field.get_field_type())  # type: ignore

        self.marshamallow = fields.List(self.field.marshamallow)  # type: ignore

        if isinstance(self.field, EmbeddedField):
            self.marshamallow = fields.List(self.field.marshamallow)

            def load(value: Any, *, partial: bool = False) -> list[FieldType]:
                return [self.field.load(item, partial=partial) for item in value]  # type: ignore

            def dump(value: list[FieldType]) -> list[dict[str, Any]]:
                return [self.field.dump(item) for item in value]  # type: ignore

            def to_mongo(value: list[FieldType]) -> list[dict[str, Any]]:
                return [self.field.to_mongo(item) for item in value]  # type: ignore

            self.load = load
            self.dump = dump
            self.to_mongo = to_mongo

    def __set_name__(self, owner: type[Document], name: str) -> None:
        if isinstance(self.field, EmbeddedField):
            self.field.schema.__lazy_init_fields__()
        return super().__set_name__(owner, name)

    def get_field_type(self) -> type[list[FieldType]]:
        return list[self.field.get_field_type()]  # type: ignore


def type_to_field(type_: type) -> Field:
    from .table import Document

    if type_ is str:
        return StringField()
    if type_ is int:
        return IntegerField()
    if type_ is float:
        return FloatField()
    if type_ is bool:
        return BooleanField()
    if type_ is datetime:
        return DateTimeField()
    if type_ is decimal.Decimal:
        return DecimalField()
    if type_ is ObjectId:
        return ObjectIdField()
    if issubclass(type_, Document):
        return EmbeddedField(type_)
    if get_origin(type_) is list:
        return ListField(type_to_field(get_args(type_)[0]))
    raise ValueError(f"Cannot convert type {type_} to field")



__all__ = ["Contains", "StartsWith", "EndsWith"]

SAFE_REGEX_DICTIONARY = {
    "(": "\\(",
    ")": "\\)",
    "!": "\\!",
    "^": "\\^",
    "$": "\\$",
    "[": "\\[",
    "]": "\\]",
    "*": "\\*",
    "+": "\\+",
    "?": "\\?",
    "{": "\\{",
    "}": "\\}",
    "/": "\\/",
    "|": "\\|",
    "<": "\\<",
    ">": "\\>",
    ".": "\\.",
    "\\": "\\\\",
}


class Contains(dict):
    def __init__(self, value: str, case_sensitive: bool = True):
        safe_value = value.translate(value.maketrans(SAFE_REGEX_DICTIONARY))
        if case_sensitive:
            super().__init__({"$regex": safe_value})
        else:
            super().__init__({"$regex": safe_value, "$options": "i"})


class StartsWith(dict):
    def __init__(self, value: str, case_sensitive: bool = True):
        safe_value = value.translate(value.maketrans(SAFE_REGEX_DICTIONARY))
        if case_sensitive:
            super().__init__({"$regex": f"^{safe_value}"})
        else:
            super().__init__({"$regex": f"^{safe_value}", "$options": "i"})


class EndsWith(dict):
    def __init__(self, value: str, case_sensitive: bool = True):
        safe_value = value.translate(value.maketrans(SAFE_REGEX_DICTIONARY))
        if case_sensitive:
            super().__init__({"$regex": f"{safe_value}$"})
        else:
            super().__init__({"$regex": f"{safe_value}$", "$options": "i"})



import dataclasses

import pytest

from typedmongo.expressions import (
    CombineExpression,
    CompareExpression,
    CompareMixin,
    NotExpression,
    RawExpression,
)


@dataclasses.dataclass(eq=False)
class Field(CompareMixin):
    field_name: str


field = Field("name")


@pytest.mark.parametrize(
    "expression, expected, compiled",
    [
        (field >= 18, CompareExpression(field, ">=", 18), {"name": {"$gte": 18}}),
        (field > 18, CompareExpression(field, ">", 18), {"name": {"$gt": 18}}),
        (field < 18, CompareExpression(field, "<", 18), {"name": {"$lt": 18}}),
        (field <= 18, CompareExpression(field, "<=", 18), {"name": {"$lte": 18}}),
        (field == "Aber", CompareExpression(field, "==", "Aber"), {"name": "Aber"}),
        (
            field != "Aber",
            CompareExpression(field, "!=", "Aber"),
            {"name": {"$ne": "Aber"}},
        ),
        (
            ~(field == "Aber"),
            NotExpression(CompareExpression(field, "==", "Aber")),
            {"name": {"$not": "Aber"}},
        ),
        (
            (field > 18) & (field < 35),
            CombineExpression(
                "AND",
                CompareExpression(field, ">", 18),
                CompareExpression(field, "<", 35),
            ),
            {"$and": [{"name": {"$gt": 18}}, {"name": {"$lt": 35}}]},
        ),
        (
            (field > 18) | (field < 35),
            CombineExpression(
                "OR",
                CompareExpression(field, ">", 18),
                CompareExpression(field, "<", 35),
            ),
            {"$or": [{"name": {"$gt": 18}}, {"name": {"$lt": 35}}]},
        ),
        (
            (field > 18) & (field < 35) | (field == 35),
            CombineExpression(
                "OR",
                CombineExpression(
                    "AND",
                    CompareExpression(field, ">", 18),
                    CompareExpression(field, "<", 35),
                ),
                CompareExpression(field, "==", 35),
            ),
            {
                "$or": [
                    {
                        "$and": [{"name": {"$gt": 18}}, {"name": {"$lt": 35}}],
                    },
                    {"name": 35},
                ]
            },
        ),
        (
            ~((field > 18) & (field < 35)),
            CombineExpression(
                "OR",
                NotExpression(CompareExpression(field, ">", 18)),
                NotExpression(CompareExpression(field, "<", 35)),
            ),
            {
                "$or": [
                    {"name": {"$not": {"$gt": 18}}},
                    {"name": {"$not": {"$lt": 35}}},
                ]
            },
        ),
        (
            ~((field > 18) | (field < 35)),
            CombineExpression(
                "AND",
                NotExpression(CompareExpression(field, ">", 18)),
                NotExpression(CompareExpression(field, "<", 35)),
            ),
            {
                "$and": [
                    {"name": {"$not": {"$gt": 18}}},
                    {"name": {"$not": {"$lt": 35}}},
                ]
            },
        ),
        (
            ~((field > 18) & (field < 35) | (field == 35)),
            CombineExpression(
                "AND",
                CombineExpression(
                    "OR",
                    NotExpression(CompareExpression(field, ">", 18)),
                    NotExpression(CompareExpression(field, "<", 35)),
                ),
                NotExpression(CompareExpression(field, "==", 35)),
            ),
            {
                "$and": [
                    {
                        "$or": [
                            {"name": {"$not": {"$gt": 18}}},
                            {"name": {"$not": {"$lt": 35}}},
                        ]
                    },
                    {"name": {"$not": 35}},
                ]
            },
        ),
        (
            ~~((field > 18) & (field < 35)),
            CombineExpression(
                "AND",
                CompareExpression(field, ">", 18),
                CompareExpression(field, "<", 35),
            ),
            {"$and": [{"name": {"$gt": 18}}, {"name": {"$lt": 35}}]},
        ),
        (
            None & (field > 18),
            CompareExpression(field, ">", 18),
            {"name": {"$gt": 18}},
        ),
        (
            (field > 18) & None,
            CompareExpression(field, ">", 18),
            {"name": {"$gt": 18}},
        ),
        (
            RawExpression({"name": "Aber"}) | (field > 18),
            CombineExpression(
                "OR",
                RawExpression({"name": "Aber"}),
                CompareExpression(field, ">", 18),
            ),
            {"$or": [{"name": "Aber"}, {"name": {"$gt": 18}}]},
        ),
    ],
)
def test_compile_expressions(expression, expected, compiled):
    assert expression == expected
    assert expression.compile() == compiled
